/* 
 * BERcut-ET. Gigabit Ethernet analyser.
 * Copyright (C) 2008 by NTC-Metrotek, SPb. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * http://metrotek.spb.ru
 * support@metrotek.spb.ru
 *
 */
#include <string.h>
#include <png/util.h>
#include <checksum/adler32.h>
#include "compress.h"

#define COMPR_METHOD_DEFLATE  8 // see rfc1950, page 4

static unsigned long gl_adler;

/*! Get 'deflate' data format header for buffer with specified len.
 *  Header is complies to rfc1951.
 *  
 *  \param len -- length of buffer which header is created for
 *  \param hdr -- buffer for header, sizeof(hdr) = 5
 *  \param final -- means that header is created for last block in sequence
 */
static void deflate_header( unsigned short len, unsigned char* hdr, 
                                                unsigned char final )
{
  // block format according to 3.2.3 section.
  //   BTYPE = 'no compression'
  //   FINAL = 1
  unsigned char block_format = final ? 1 : 0;

  // one's complement of len, see 3.2.4 section
  unsigned short nlen = 0xffff - len;

  hdr[0] = block_format;
  memcpy( &hdr[1], &len, sizeof(len) );
  memcpy( &hdr[3], &nlen, sizeof(nlen) );
}
                 
/*! \brief Write zlib header complied with rfc1950 to 'hdr'.
 *         sizeof(hdr) = 2;
 */
static void zlib_header( unsigned char* hdr )
{
  unsigned char cm = COMPR_METHOD_DEFLATE;
  unsigned char cinfo = 7; // 32K window size (maximum)

  // compression method and flags
  unsigned char cmf = (cm & 0xf) | (cinfo << 4);

  // FLaGs
  // static value 1 means:
  //   * check bits for CMF and FLG.
  //   * compression level = 0 (fastest = store)
  //   * not use dictionaries
  // see rfc1950, page 4, FLG.
  unsigned char flg = 1;
  
  hdr[0] = cmf;
  hdr[1] = flg;
}


/*! \brief Compress buffer with 'deflate' algorithm (rfc1951)
 *         and pack it to zlib format (rfc1950).
 *         Compression level = 0 (store).
 *
 *         in (len):
 *                                    +------------------+
 *                                    |       data       |
 *                                    |                  |
 *                                    +------------------+
 *
 *         out (len + 11)
 *         +----------+---------------+------------------+------------------+
 *         | zlib hdr | 'deflate' hdr | data w/o changes | adler32(data)    |
 *         | 2 bytes  |    5 bytes    |     len          |     4 bytes      |
 *         +----------+---------------+------------------+------------------+
 *
 *         \param final -- means that yacompress is called for last 
 *                         block. This flag is used by deflate algorithm.
 *                         1 -- final.
 *         \param in -- input buffer. if (in == NULL) then 
 *                      initialization performed
 *
 *         \return the count of compressed data in 'out'. can't be more than
 *                 'len + 11'.
 */
unsigned short yacompress( unsigned char* in, unsigned short len,
                           unsigned char* out, unsigned char final )
{
  static char first = 1;
  if (in == NULL) {
    gl_adler = adler32( 0, NULL, 0 );
    first = 1;
    return 0;
  }

  unsigned char* data = out;

  if (first) {
    zlib_header( data );
    data += 2;
    first = 0;
  }

  deflate_header( len, data, final );
  data += 5;

  memcpy( data, in, len );
  gl_adler = adler32( gl_adler, in, len );
  data += len;

  if (final) {
    save_uint_32( data, gl_adler );
    data += sizeof( gl_adler );
  }

  return (unsigned short)(data - out);
}

